@@ -2,7 +2,7 @@
  *
  * rasqal_raptor.c - Rasqal triple store implementation with raptor
  *
- * $Id: rasqal_raptor.c 11565 2006-10-31 05:15:22Z dajobe $
+ * $Id: rasqal_raptor.c 13196 2007-11-14 07:48:10Z laalto $
  * 
  * Copyright (C) 2004-2006, David Beckett http://purl.org/net/dajobe/
  * Copyright (C) 2004-2005, University of Bristol, UK http://www.bristol.ac.uk/
@@ -51,6 +51,8 @@ struct rasqal_raptor_triple_s {
 typedef struct rasqal_raptor_triple_s rasqal_raptor_triple;
 
 typedef struct {
+  rasqal_query* query;
+
   rasqal_raptor_triple *head;
   rasqal_raptor_triple *tail;
 
@@ -62,11 +64,16 @@ typedef struct {
   /* size of the two arrays below */
   int sources_count;
   
-  /* array of shared pointers into query->data_graph uris */
-  raptor_uri **source_uris;
+  /* shared pointers into query->data_graph uris */
+  raptor_uri* source_uri;
 
   /* array of URI literals (allocated here) */
   rasqal_literal **source_literals;
+
+  /* genid base for mapping user bnodes */
+  unsigned char* mapped_id_base;
+  /* length of above string */
+  int mapped_id_base_len;
 } rasqal_raptor_triples_source_user_data;
 
 
@@ -209,6 +216,31 @@ rasqal_raptor_error_handler(void *user_data,
 }
 
 
+static unsigned char*
+rasqal_raptor_generate_id_handler(void *user_data,
+                                  raptor_genid_type type,
+                                  unsigned char *user_bnodeid) 
+{
+  rasqal_raptor_triples_source_user_data* rtsc=(rasqal_raptor_triples_source_user_data*)user_data;
+  if(user_bnodeid) {
+    unsigned char *mapped_id;
+    mapped_id=(unsigned char*)RASQAL_MALLOC(cstring, 
+                                            rtsc->mapped_id_base_len + 1 + 
+                                            strlen((const char*)user_bnodeid) + 1);
+    strncpy((char*)mapped_id,  (const char*)rtsc->mapped_id_base, 
+            rtsc->mapped_id_base_len);
+    mapped_id[rtsc->mapped_id_base_len]='_';
+    strcpy((char*)(mapped_id+rtsc->mapped_id_base_len+1),
+           (const char*)user_bnodeid);
+
+    raptor_free_memory(user_bnodeid);
+    return mapped_id;
+  }
+  
+  return rasqal_query_get_genid(rtsc->query, (const unsigned char*)"genid", -1);
+}
+
+
 static int
 rasqal_raptor_new_triples_source(rasqal_query* rdf_query,
                                  void *factory_user_data,
@@ -230,20 +262,29 @@ rasqal_raptor_new_triples_source(rasqal_query* rdf_query,
   if(!rtsc->sources_count)
     return -1;  /* no data */
 
-  rtsc->source_uris=(raptor_uri**)RASQAL_CALLOC(raptor_uri_ptr, rtsc->sources_count, sizeof(raptor_uri*));
   rtsc->source_literals=(rasqal_literal**)RASQAL_CALLOC(rasqal_literal_ptr, rtsc->sources_count, sizeof(rasqal_literal*));
 
+  rtsc->query=rdf_query;
+
   for(i=0; i< rtsc->sources_count; i++) {
     rasqal_data_graph *dg=(rasqal_data_graph*)raptor_sequence_get_at(rdf_query->data_graphs, i);
     raptor_uri* uri=dg->uri;
+    raptor_uri* name_uri=dg->name_uri;
 
     rtsc->source_index=i;
-    rtsc->source_uris[i]=raptor_uri_copy(uri);
-    rtsc->source_literals[i]=rasqal_new_uri_literal(raptor_uri_copy(uri));
+    rtsc->source_uri=raptor_uri_copy(uri);
+    if(name_uri)
+      rtsc->source_literals[i]=rasqal_new_uri_literal(raptor_uri_copy(name_uri));
+    rtsc->mapped_id_base=rasqal_query_get_genid(rdf_query,
+                                                (const unsigned char*)"graphid",
+                                                i);
+    rtsc->mapped_id_base_len=strlen((const char*)rtsc->mapped_id_base);
 
     parser=raptor_new_parser("guess");
     raptor_set_statement_handler(parser, rtsc, rasqal_raptor_statement_handler);
     raptor_set_error_handler(parser, rdf_query, rasqal_raptor_error_handler);
+    raptor_set_generate_id_handler(parser, rtsc,
+                                   rasqal_raptor_generate_id_handler);
 
 #ifdef RAPTOR_FEATURE_NO_NET
     if(rdf_query->features[RASQAL_FEATURE_NO_NET])
@@ -253,6 +294,12 @@ rasqal_raptor_new_triples_source(rasqal_query* rdf_query,
 
     raptor_parse_uri(parser, uri, dg->name_uri);
     raptor_free_parser(parser);
+
+    raptor_free_uri(rtsc->source_uri);
+    /* This is freed in rasqal_raptor_free_triples_source() */
+    /* rasqal_free_literal(rtsc->source_literal); */
+    RASQAL_FREE(cstring, rtsc->mapped_id_base);
+    
     if(rdf_query->failed) {
       rasqal_raptor_free_triples_source(user_data);
       break;
@@ -267,7 +314,7 @@ rasqal_raptor_new_triples_source(rasqal_query* rdf_query,
  * rasqal_raptor_triple_match:
  * @triple: #rasqal_triple to match against
  * @match: #rasqal_triple with wildcards
- *
+ * @parts; parts of the triple to match
  * .
  * 
  * Match a rasqal_triple against a rasqal_triple with NULL
@@ -276,30 +323,69 @@ rasqal_raptor_new_triples_source(rasqal_query* rdf_query,
  * Return value: non-0 on match
  **/
 static int
-rasqal_raptor_triple_match(rasqal_triple *triple, rasqal_triple *match) {
-  if(match->subject) {
-    if(!rasqal_literal_equals(triple->subject, match->subject))
-      return 0;
+rasqal_raptor_triple_match(rasqal_triple *triple, rasqal_triple *match,
+                           rasqal_triple_parts parts)
+{
+  int rc=0;
+  
+#if RASQAL_DEBUG > 1
+  RASQAL_DEBUG1("\ntriple ");
+  rasqal_triple_print(triple, stderr);
+  fputs("\nmatch  ", stderr);
+  rasqal_triple_print(match, stderr);
+  fputs("\n", stderr);
+#endif
+
+  if(match->subject && (parts & RASQAL_TRIPLE_SUBJECT)) {
+    if(!rasqal_literal_equals_flags(triple->subject, match->subject,
+                                    RASQAL_COMPARE_RDF, NULL))
+      goto done;
   }
 
-  if(match->predicate) {
-    if(!rasqal_literal_equals(triple->predicate, match->predicate))
-      return 0;
+  if(match->predicate && (parts & RASQAL_TRIPLE_PREDICATE)) {
+    if(!rasqal_literal_equals_flags(triple->predicate, match->predicate,
+                                    RASQAL_COMPARE_RDF, NULL))
+      goto done;
   }
 
-  if(match->object) {
-    if(!rasqal_literal_equals(triple->object, match->object))
-      return 0;
+  if(match->object && (parts & RASQAL_TRIPLE_OBJECT)) {
+    if(!rasqal_literal_equals_flags(triple->object, match->object,
+                                    RASQAL_COMPARE_RDF, NULL))
+      goto done;
+  }
+
+  if(parts & RASQAL_TRIPLE_ORIGIN) {
+    /* Binding a graph */
+    
+    /* If expecting a graph and triple has none then no match */
+    if(!triple->origin)
+      goto done;
+      
+    if(match->origin) {
+      if(match->origin->type == RASQAL_LITERAL_URI ) {
+        raptor_uri* triple_uri=triple->origin->value.uri;
+        raptor_uri* match_uri=match->origin->value.uri;
+        if(!raptor_uri_equals(triple_uri, match_uri))
+          goto done;
+      }
+    }
+    
+  } else {
+    /* Not binding a graph */
+    
+    /* If triple has a GRAPH and there is none in the triple pattern, no match */
+    if(triple->origin)
+      goto done;
   }
   
-  if(match->origin && match->origin->type == RASQAL_LITERAL_URI ) {
-    raptor_uri* triple_uri=triple->origin->value.uri;
-    raptor_uri* match_uri=match->origin->value.uri;
-    if(!raptor_uri_equals(triple_uri, match_uri))
-      return 0;
-  }
+  rc=1;
+  done:
   
-  return 1;
+#if RASQAL_DEBUG > 1
+  RASQAL_DEBUG2("result: %s\n", (rc ? "match" : "no match"));
+#endif
+
+  return rc;
 }
 
 
@@ -310,9 +396,13 @@ rasqal_raptor_triple_present(rasqal_triples_source *rts, void *user_data,
 {
   rasqal_raptor_triples_source_user_data* rtsc=(rasqal_raptor_triples_source_user_data*)user_data;
   rasqal_raptor_triple *triple;
+  rasqal_triple_parts parts=RASQAL_TRIPLE_SPO;
+  
+  if(t->origin)
+    parts |= RASQAL_TRIPLE_GRAPH;
 
   for(triple=rtsc->head; triple; triple=triple->next) {
-    if(rasqal_raptor_triple_match(triple->triple, t))
+    if(rasqal_raptor_triple_match(triple->triple, t, parts))
       return 1;
   }
 
@@ -336,14 +426,10 @@ rasqal_raptor_free_triples_source(void *user_data) {
   }
 
   for(i=0; i< rtsc->sources_count; i++) {
-    if(rtsc->source_uris[i])
-      raptor_free_uri(rtsc->source_uris[i]);
     if(rtsc->source_literals[i])
       rasqal_free_literal(rtsc->source_literals[i]);
   }
-  RASQAL_FREE(raptor_uri_ptr, rtsc->source_uris);
   RASQAL_FREE(raptor_literal_ptr, rtsc->source_literals);
-  
 }
 
 
@@ -360,6 +446,9 @@ typedef struct {
   rasqal_raptor_triple *cur;
   rasqal_raptor_triples_source_user_data* source_context;
   rasqal_triple match;
+
+  /* parts of the triple above to match: always (S,P,O) sometimes C */
+  rasqal_triple_parts parts;
 } rasqal_raptor_triples_match_context;
 
 
@@ -392,8 +481,9 @@ rasqal_raptor_bind_match(struct rasqal_triples_match_s* rtm,
 
   if(bindings[1] && (parts & RASQAL_TRIPLE_PREDICATE)) {
     if(bindings[0] == bindings[1]) {
-      if(rasqal_literal_compare(rtmc->cur->triple->subject,
-                                rtmc->cur->triple->predicate, 0, &error))
+      if(!rasqal_literal_equals_flags(rtmc->cur->triple->subject,
+                                      rtmc->cur->triple->predicate,
+                                      RASQAL_COMPARE_RDF, &error))
         return (rasqal_triple_parts)0;
       if(error)
         return (rasqal_triple_parts)0;
@@ -411,8 +501,9 @@ rasqal_raptor_bind_match(struct rasqal_triples_match_s* rtm,
     int bind=1;
     
     if(bindings[0] == bindings[2]) {
-      if(rasqal_literal_compare(rtmc->cur->triple->subject,
-                                rtmc->cur->triple->object, 0, &error))
+      if(!rasqal_literal_equals_flags(rtmc->cur->triple->subject,
+                                      rtmc->cur->triple->object,
+                                      RASQAL_COMPARE_RDF, &error))
         return (rasqal_triple_parts)0;
       if(error)
         return (rasqal_triple_parts)0;
@@ -423,8 +514,9 @@ rasqal_raptor_bind_match(struct rasqal_triples_match_s* rtm,
     if(bindings[1] == bindings[2] &&
        !(bindings[0] == bindings[1]) /* don't do this check if ?x ?x ?x */
        ) {
-      if(rasqal_literal_compare(rtmc->cur->triple->predicate,
-                                rtmc->cur->triple->object, 0, &error))
+      if(!rasqal_literal_equals_flags(rtmc->cur->triple->predicate,
+                                      rtmc->cur->triple->object,
+                                      RASQAL_COMPARE_RDF, &error))
         return (rasqal_triple_parts)0;
       if(error)
         return (rasqal_triple_parts)0;
@@ -461,7 +553,7 @@ rasqal_raptor_next_match(struct rasqal_triples_match_s* rtm, void *user_data)
   while(rtmc->cur) {
     rtmc->cur=rtmc->cur->next;
     if(rtmc->cur &&
-       rasqal_raptor_triple_match(rtmc->cur->triple, &rtmc->match))
+       rasqal_raptor_triple_match(rtmc->cur->triple, &rtmc->match, rtmc->parts))
       break;
   }
 }
@@ -546,6 +638,7 @@ rasqal_raptor_init_triples_match(rasqal_triples_match* rtm,
 
   m->bindings[2]=var;
   
+  rtmc->parts = RASQAL_TRIPLE_SPO;
 
   if(t->origin) {
     if((var=rasqal_literal_as_variable(t->origin))) {
@@ -554,11 +647,12 @@ rasqal_raptor_init_triples_match(rasqal_triples_match* rtm,
     } else
       rtmc->match.origin=rasqal_new_literal_from_literal(t->origin);
     m->bindings[3]=var;
+    rtmc->parts |= RASQAL_TRIPLE_GRAPH;
   }
   
 
   while(rtmc->cur) {
-    if(rasqal_raptor_triple_match(rtmc->cur->triple, &rtmc->match))
+    if(rasqal_raptor_triple_match(rtmc->cur->triple, &rtmc->match, rtmc->parts))
       break;
     rtmc->cur=rtmc->cur->next;
   }
